\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\\
\\
\\  Coppersmith - Howgrave-Graham Primality Prover
\\  ----------------------------------------------
\\
\\
\\
\\  Author:  John Renze
\\      with help from David Broadhurst, Greg Childers and the 
\\      PrimeForm community.
\\
\\  Version:  0.1
\\
\\  Description:  This GP script completes the search for divisors in
\\     residue classes required by BLS in cases where n^2-1 is factored
\\     to more than 25% of n.  It is based on the algorithm of Coppersmith 
\\     and Howgrave-Graham using lattice basis reduction.  More description
\\     of the algorithms involved is given in the documentation for the
\\     various sections of code.
\\
\\  Features:  This code is loosely based on the original GP script by 
\\    Greg Childers and Mathematica code by John Renze with numerous 
\\    enhancements and modifications.  In addition to some general tidying 
\\    up, it includes the following features of importance.
\\    - Includes right endpoint modification due to Broadhurst for the case
\\      F >> G.  Also includes the mirror image right endpoint modification
\\      for the case G >> F.
\\    - Works from right-to-left in the interval, instead of right-to-left.
\\      This ensures that we do not overshoot the right endpoint and wind up
\\      doing too much work in the hardest part of the interval.  The formula
\\      for the selection of u is derived for this new situation.
\\    - Implements better heuristics for the selection of h-values.  This
\\      could still use some work, though.
\\    And most importantly...
\\    - Produces and certificates for each of the CHG intervals.
\\      These certificates are saved to file and become part of the
\\      primality proof of any input.
\\
\\  Todo List:
\\    - Certificate verification.
\\    - Faster and better h selection code.  I like the values of h I am
\\      getting.  It's just that the way I'm producing them is slightly absurd.
\\    - Interruptiblility.  My idea was to save each interval certificate to
\\      a temporary file, so that we would never lose more than the interval
\\      we were working on.
\\    - Client/Server.
\\    - Explore using the Fincke-Pohst algorithm to further push the lower
\\      endpoint of each subinterval.
\\
\\  Usage:  Follow the directions below for the creation of an input file and
\\    then just run the whole thing.  This script should Just Work for 
\\    factored portions of more than about 28% on inputs of up to 10,000 
\\    digits or so and should take no more than a few hours.  If you want
\\    to push further than that, particularly with regards to smaller
\\    factorization percentages, you will probably have to change some
\\    parameters.
\\
\\  Landmarks:
\\    Used at 30.3% on fibonacci(37511) by John Renze
\\    Used at 27.8% on (127^5281-1)/126 by Predrag Minovic
\\
\\
\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

allocatemem(64*1024*1024);   \\ Increase stack to 64mb for now

\\  Precision:  There are several places you have to take roots and all the digits
\\    matter.  Unfortunately, you're going to have to set this yourself.  If it is 
\\    too high, the h selection code runs too slowly.  If it is too low, you will 
\\    get precision errors.  Here are some guidelines and observations based on
\\    past use.
\\
\\  \p100 works at 500 digits
\\  \p200 works at 1000 digits
\\  \p1500 works for 7800 digits and 30%
\\  \p4000 works for 10,000 digits and 28%
\\  \p5000 works for 11,109 digits and 27.8%

\p4000


\\  PROGRAM INPUT
\\  -------------

worktodofile="2879_2851.in";
certificatefile="2879_2851.out";

\\  The input to the CHG program should be stored in a GP-readable file containing
\\  three definitions:
\\    - n is the number being factored.
\\    - F is the factored portion of n-1.
\\    - G is the factored portion of n+1.
\\  The factored portions should be given as numbers, not lists of factors.
\\  There is currently no functionality for queuing multiple jobs in the same
\\  todo file.
\\
\\  If you wish to simply verify certificates, make a worktodo file containing
\\  the single line:
\\     verify=1;
\\  This will verify the last certificate found in the file certificatefile.
\\  WARNING:  No error checking is done to see if the last expression in
\\    certificatefile actually is a CHG certificate.
\\
\\ Sample input file:
\\
\\ n=fibonacci(37511);
\\ F = 62775143316859388841343638603760262885355327571476168714579513231335302613\
\\ 394722737516948697279015446872244680852417335389062567672892834443266967227044\
\\ 267343107602566591787279020338249659594775543877116273589325495771559934273871\
\\ 736198959911039262976220481365049494798582557142393506255394560426252984397134\
\\ 264805559126067594414685622039416294414069197324383657636212099234154712846489\
\\ 128254512947642777898134848995030480373224043573839670432336205167893646569600\
\\ 503890523663852950826444728875972322721433289449049879465641682599071191381196\
\\ 756527227045053978427772432650142064521751543981912567101774463940191510554041\
\\ 967441498909777385119996089131805819704236144019042950223066429834658810755952\
\\ 915594974085917255631336585819067483496626136725501140976318827626852710907733\
\\ 428296966075921655223486923766334592371345873384609771413963234629052885619137\
\\ 198484422933419383905440566921061058547239077830745550376853825143554269068029\
\\ 449416713312565610280055795292302517509520424094327264150437091075361474147563\
\\ 716120579418817904034751446149777388568589845925632099815267659046338282003777\
\\ 324778080088345745233390488131469329738992659480732239690650664051608303653461\
\\ 416009856172648669796174888917258833432565723176581025547412817910723213079894\
\\ 070277099330861173024311289318672222771048826111722474732910513476917806698754\
\\ 704343377676181835471828821132041187576373731476736593861232270460509798014163\
\\ 212789477392005224436722801716554526581223823253954775720084105973712309688932\
\\ 527202080976782466194314070923463911629667647443115709289214536710249673191670\
\\ 565181760250273459330680656893555121034823843096956951657856786474281312682610\
\\ 659781681219315578335436607392721451281118743107596250637610736059150491398703\
\\ 5729509151465702572889867607291636625838251472069418997008;
\\ G = 11648808483349882397703211029523510905183282015824218280705166091134338009\
\\ 411325933592619625636656177820586478393288822421659254761363726413083385587338\
\\ 867518516214971681804184589436399993469990459824077732123711420425610612394567\
\\ 712968952615033175001882527699192847945461631935784896776570494554271148948930\
\\ 201212709833891180658622734257147004240879716986931736167218984732477945702269\
\\ 024971673439732881532959994970816752180978833038764952552957556454527707489141\
\\ 241912860677792131306991659345145332075753404299832748735969589464010284887771\
\\ 613711187365732672449406096951515112284586903161087975680530346370;

\\  ADJUSTIBLE PARAMETERS
\\  ---------------------
\\
\\      Only change these if you are having problems or are really stretching
\\      the abilities of the program.  the defaults should work fine for a >= 28%
\\      and candidates of around 10,000 digits or less.
\\
\\  Maximum size of matrices to use.  For factored portions of 28% or more, 
\\  this can be somewhere in the neighborhood of 8-12.  Don't be alarmed if
\\  the program actually uses a smaller value.
maxh = 12;
\\  Minimum h to try.  This is to keep the program from trying silly
\\  values of h.  You probably won't need to change this, although if you
\\  keep getting precision errors during the search for h, you can try
\\  raising it.
minh = 5;
\\  Maximum number of intervals to use when choosing h-values.  You can set this to something
\\  pretty big.  This is just here to keep the code that chooses h from
\\  spending ridiculous amounts of time on h's that are too small.
maxintervals = 20;

\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\\ Code - No user modifiable parameters below this line.
\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

\\***********************************************
\\	Certificates code
\\***********************************************
\\

\\ A certificate is a collection of data which allows the factoring for
\\ a particular subinterval to be (provably) recreated without redoing
\\ the LLL step.  This is important because it allows primality
\\ proofs to be independently checked without rerunning the whole
\\ algorithm with a different implementation.
\\
\\ A CHG certificate is a list with 4 parts:
\\ [n,s,list1,list2], where each part of list1 is a list of certificates 
\\ for the subintervals in the search congruent to 1 mod s, and each part 
\\ of list2 is a list of certificates for the subintervals in the search 
\\ congruent to n mod s.
\\
\\ The certificate for each subinterval is a list containing five 
\\ elements.  They are
\\
\\    1)  The value of h for that interval.
\\    2)  The value of u for that interval.
\\    3)  The left endpoint of the interval.
\\    4)  The right endpoint of the interval.
\\    5)  Polynomial corresponding to the short vector.
\\
\\ Remark:  This is different from the certificate that David and I had previously
\\   discussed and agreed upon.  Why?  The polynomial is much shorter than the 
\\   weight vector.  I was becoming concerned that these certificates were getting
\\   a bit large.  The objection to using using that polynomial is that you need to
\\   invert the matrix M to verify that the polynomial is an integer linear combination
\\   of the original polynomials.  However, M is lower triangular, so this is
\\   less expensive than originally thought.

certificateCreateInterval (h, u, X, Y, poly) = 
{
    local (certificate);
    
    certificate = vector(5);
    
    certificate[1] = h;
    certificate[2] = u;
    certificate[3] = X;
    certificate[4] = Y;
    certificate[5] = poly;
    
    return (certificate);
}

certificateSave (certificate) = 
{
    print ("A certificate has been saved to the file \"", certificatefile, "\".");
    write (certificatefile, certificate);
}

\\  These functions verify  a certificate in the above form.  They return
\\  1 if the certificate is valid, 0 if it is not.  These functions are
\\  only used if the verify flag is set in the worktodo file.
certificateVerifyInterval (certificate) = 
{
    \\ Not yet implemented
}

certificateVerify (certificate) =
{
    \\ Not yet implemented
}

\\***********************************************
\\	CHG code
\\***********************************************

\\ This formula for the value of u is derived by taking the deriviative
\\ of the lower limit with respect to u and setting it equal to zero.
\\ Q: Is it better to round up or down?
bestu (n, h, Y) =
{
    return (max (1, round (sqrt (h*(h-1)*log(Y)/log(n))))); 
}


\\  We choose h by running "simulations" of the CHG process.  Essentially, I fix
\\  h and then iterate the LLL condition until I get a rightendpoint of 1.  I do
\\  this for every value between minh and maxh.  The chosen value of h is the one 
\\  that minimizes (number of intervals)*h^4.  Because we are working at such high
\\  precision, this takes a bit of time, but I think it is worth it because even 
\\  small reductions in h will result in large time savings.
\\  Q:  I would prefer to do the h selection at a lower precision than the
\\      rest of the calculation.  How do I tell PARI that I don't care if I 
\\      lose a few digits?
chooseh (n, s, r, leftendpoint, Y) =
{
    local (h, u, YY, numintervals, currweight, bestweight, besth, starth);
    
    besth = maxh;
    bestweight = maxintervals*maxh^4;
       
    for (h = minh, maxh,
        YY = Y;
	numintervals = 0;   
        while(YY >= leftendpoint,
	    u = bestu (n, h, YY);
            numintervals += 1;
            if (numintervals >= maxintervals,
                numintervals = 0;
                break();
            );
            YY = floor((h^(1/2)*2^((h - 1)/4)*YY^((h - 1)/2)*n^(u*(u + 1)/(2*h)))^(1/u)/s);
        );
        currweight = numintervals * h^4;
        if (currweight != 0 && bestweight > currweight,
            bestweight = currweight;
            besth = h
        );
    );

    return (besth);
}

\\  This is the routine which does all of the heavy lifting.
\\  Pre:  The values of n, s, r and given.  We also give the parameters h, u, and Y.
\\        No error checking is done to see if these parameter choices make sense.
\\  Post: The first factor found is returned.  If no factor is found, a certificate
\\        attesting to the absence of factors is returned.
CHG (n, s, r, h, u, Y) = 
{
    local(m, X, poly, M, reducedbasis, shortvector, p, newpoly, f, i, j, certificate, newshortvector, YY);
   
    print ("    Running CHG with h = ", h, ", u = ", u, ". Right endpoint has ", length(Str(Y)), " digits.");   

    \\ Set up matrix for LLL.
    m=(r*((1/s)%n))%n;
    M=matrix(h,h);
    for(i=0,u,
        poly = n^(u-i)*(x+m)^i;
        for(j=0,h-1,
           M[i+1,j+1]=polcoeff(poly,j)*Y^j));
    for(i=u+1,h-1,
        poly = x^(i-u)*(x+m)^u;
        for(j=0,h-1,
           M[i+1,j+1]=polcoeff(poly,j)*Y^j));

    \\ Perform LLL and assemble resulting polynomial.
    \\ The floating point algorithm is used.  In the unlikely event that there
    \\ is an error, it will be detected by the certification code.
    reducedbasis=(M~*qflll(M~))~;
    shortvector = vector(h,i,reducedbasis[1,i]);
    X = ceil ((h*norml2(shortvector))^(1/(2*u))/s);
    newpoly=0;
    for(j=0,h-1,newpoly=newpoly + shortvector[j+1]*(z/Y)^j);

    \\ Search for integer roots of the resulting polynomial.    
    \\ We use factor here.  If we by extremely bad luck get a polynomial
    \\ which is hard to factor, we can use an algorithm which is 
    \\ provable polynomial time such as Sturm sequences.
    f=factor(newpoly)[,1]~;
    for(j=1,length(f),
        p=f[j];
        if( poldegree(p)==1 && pollead(p)==1, return(s*(z-p)+r)));

    \\ The same polynomial can be used for several subintervals.  These subintervals
    \\ are smaller than those we get if we redo the LLL, but since they
    \\ are free, we may as well take them.  The verifier knows to look for multiple
    \\ intervals, so we just give the left endpoint of the whole string of them
    \\ in the certificate.
    YY = Y;            \\ The 10 is there because we don't bother computing a 
    while (10*X < YY,  \\ new left endpoint if we can't get at least one more digit.
        YY = X - 1;
        newshortvector = vector (h, i, polcoeff(newpoly,i)*YY^i);
        X = ceil ((h*norml2(newshortvector))^(1/(2*u))/s)
    );

    \\ Create certificate and return it.
    certificate = certificateCreateInterval (h, u, X, Y, newpoly);
    return (certificate);
}


\\ The main CHG routine.  This is the only function is this section which
\\ should be called from elsewhere.
\\ Pre:   It is assumed that s^4 > n.  This is not checked.
\\ Post:  Returns the first factor of n found of the form s x + r 
\\        with leftendpoint <= x <= rightendpoint.  After the first factor is
\\        found, the remaining cofactor qualifies for Lenstra's algorithm,
\\        so the CHG procedure should be aborted anyway.  Returns a list of 
\\        subinterval certificates if no factor is found.
CHGFactor (n, s, r, leftendpoint, rightendpoint) =
{
    local(h, u, Y, possiblefactor, certificatelist);
   
    possiblefactor = 0;
    certificatelist = listcreate(2*maxintervals); 
    \\ Allow some extra space for some low h intervals at the end.
    Y = rightendpoint;

    \\ There are so many special cases if r divides n,
    \\ let's just get that one out of the way.
    if (r > 1 && n % r == 0, return (r));
    
    while(Y >= leftendpoint,
        h = chooseh (n, s, r, leftendpoint, Y);
        u = bestu (n, h, Y);
        possiblefactor = CHG(n, s, r, h, u, Y);
        \\ possiblefactor is either a factor or the new right endpoint.
        if (type(possiblefactor) == "t_POL",
            return (possiblefactor));
        listput (certificatelist, possiblefactor);
        Y = possiblefactor[3] - 1;
    );
    
    return (certificatelist);
}

\\***********************************************
\\	Testing code
\\***********************************************

\\ This tests the CHG code. It produces a number n with d digits,
\\ a modulus s with log(s)/log(n) = a, and runs the CHG program.
\\ This function doesn't actually get called when running in
\\ production mode, but I keep it around because it is convenient
\\ for debugging the CHG portions.
tst(a,d) =
{
   local(n,r,s,c,f,g);
   s=137+random(10^(ceil(d*a)));
   until(gcd(r,s)==1,
      r=1+random(s-2));
   until(gcd(n,s)==1,
       c=137+random(10^random(ceil(d*(1/2-a))));
       f=c*s+r;
       g=137+ceil(random(10^d)/f);n=f*g);
   print("percent="round(100*log(s)/log(n)));
   print("digits="ceil(log(n)/log(10)));
   print("factor="f);
   print("n="n);
   print("s="s);
   print("r="r);
   CHGFactor(n,s,r,1,floor(sqrt(n)/s));
}

\\***********************************************
\\	Right endpoint code
\\***********************************************
\\
\\ We know from the BLS test that all factors of n must be congruent to 1
\\ or n modulo FG.  We want to find upper bounds on
\\ the size of x in order to reduce the workload for CHG. Which upper
\\ bound we use will depend on whether F or G is larger.
\\ 
\\ Bound #1: x < n^(1/2) / (F*G)
\\ - Conditions: none
\\ - Proof: The smaller factor must be smaller than sqrt(n).
\\ 
\\ Bound #2: x < (2*n)/(F^3*G^2)
\\ - Conditions: Suppose
\\       n = (F*G*y + 1))(F*G*z + (F*r + 1))
\\ where r = ((n-1)/F) % G.
\\ Write (n - (F*r + 1))/(F*G) = a*F + b where
\\ 0 <= b < F. Then (G*b + r)^2 - 4*G*a must not be a perfect square.
\\ - Proof: By construction, the quantity y + z must be congruent to b
\\ modulo F. The square test rules out the smallest possible value for
\\ this quantity.  Let x = min (y, z), w = max (y, z).  We get 2*w >
\\ y + z > F. Then
\\ x < n/(F^2*G^2*w) < (2*n)/(F^3*G^2).
\\ 
\\ Bound #3: x < (2*n)/(F^2*G^3)
\\ - Conditions: Suppose
\\       n = (F*G*y + 1))(F*G*z + (G*r - 1))
\\ where r = ((n+1)/G) % F.
\\ Write (n - (G*r - 1))/(F*G) = a*G + b where
\\ 0 <= b < G. Then (F*b + r)^2 + 4*F*a must not be a perfect square.
\\ - Proof: By construction, the quantity y + z must be congruent to b
\\ modulo G. The square test rules out the smallest possible value for
\\ this quantity.  Let x = min (y, z), w = max (y, z).  We get 2*w >
\\ y + z > G. Then
\\ x < n/(F^2*G^2*w) < (2*n)/(F^2*G^3).
\\ 
\\ Use whichever bound is smallest (of course!) and proceed with CHG.

\\ Square test to use if F >> G
squaretest1 (n, F, G) = 
{
    local(R, a, b, r, S);
    
    r = ((n-1)/F) % G;    
    R = (n - (F*r + 1))/(F*G);
    a = floor(R/F);
    b = R - a*F; 
    S = (G*b + r)^2 - 4*G*a; 
    
    return (issquare(S));
}

\\ Square test to use if G >> F
squaretest2 (n, F, G) = 
{
    local(R, a, b, r, S);
    
    r = ((n+1)/G) % F;    
    R = (n - (G*r - 1))/(F*G);
    a = floor(R/F);
    b = R - a*G; 
    S = (F*b + r)^2 - 4*F*a; 

    return (issquare(S));
}

findrightendpoint (n, F, G) =
{
    local (rightendpoint);

    if (4*n < (F^4)*(G^2), 							
	if (squaretest1 (n, F, G),
	    print ("Square test failed. n is composite."); quit(),
	    print ("Square test passed for F >> G.  Using modified right endpoint.");
            rightendpoint = floor(2*n / ((F^3)*(G^2)))
    	),
        if (4*n < (G^4)*(F^2),
            if (squaretest2 (n, F, G),
                print ("Square test failed. n is composite."); quit(),
                print ("Square test passed for G >> F.  Using modified right endpoint.");
                rightendpoint = floor(2*n / ((G^3)*(F^2)))),
            print ("Unmodified right endpoint will be used.");
            rightendpoint = floor(sqrt(n)/(F*G))
	)
    );

    return (rightendpoint);
}

\\***********************************************
\\	Left endpoint code
\\***********************************************
\\
\\ We know from the BLS test that all factors of n must be congruent to 1
\\ or n modulo FG.  We want to find lower bounds on
\\ the size of x in order to reduce the workload for CHG. 
\\ These bounds are based on the KP method.

\\print("D Broadhurst, KP combined test, 8 Jul 2001");
\\print("Hard coded for Fib(25561) - please hand edit!");

\\N=fibonacci(25561);
\\F=prod(k=1,length(lsm),lsm[k]);
\\G=prod(k=1,length(lsp),lsp[k]);
\\R=(N-1)/F;
\\S=(N+1)/G;

\\{if(denominator(R)==1
\\  &&denominator(S)==1
\\  &&gcd(F,R)==1
\\  &&gcd(F,G)==1
\\  &&F^(1/3)/6>G-1
\\  &&G-1>3*N/F^(10/3),
\\  ok=1,
\\  print("Fail");ok=0);}

\\c1=R%F;
\\c4=(R-c1)/F;
\\tG=c4%G;
\\print(tG);

\\print();

\\if(issquare((c1+tG*F)^2+4*(tG-c4)),print("Fail "tG);ok=0);

\\b=contfrac(c1/F);
\\v=0;vn=1;u=1;un=b[1];n=1;
\\{while(F^(1/3)>vn,n=n+1;bn=b[n];uo=u;u=un;vo=v;v=vn;
\\ un=bn*un+uo;vn=bn*vn+vo);}

\\d=round(c4*v/F);
\\zs=[v,u*F-c1*v,c4*v-d*F+u,-d];
\\q=0;
\\for(k=1,4,print(zs[k]);q=q+zs[k]*x^(4-k));

\\print();

\\\p1500
\\ps=polroots(q);

\\{for(k=1,3,r=floor(real(ps[k]));print(r);
\\ x=r;q1=eval(q);x=r+1;q2=eval(q);
\\ if(q1*q2>=0,print("Fail root ",k);ok=0));}

\\if(ok==1,print("OK"),print("not yet"));



\\ Performs a cube test with v < L <= v'
cubetest (n, F, L) =
{
}

\\ For now we'll just put the answer in here.  But later
findleftendpoint (n, F, G) =
{
    if (G == 1, return (ceil(F^(1/3)/4)), return (1))
}

\\***********************************************
\\	Execution starts here
\\***********************************************

print ("\nWelcome to the CHG primality prover!");
print ("------------------------------------\n");

\\ Read data in from job file.
read(worktodofile);

if (verify == 1, \
    certificateVerify( read(certificatefile) ); \
    quit());

s = gcd (n^2-1, F*G);

\\ F and G have a common factor of 2. We need to get rid of it.
\\ so that s = F*G.
if (F > G, G = G / gcd(F, G));
if (G > F, F = F / gcd(F,G));

print ("Found values of n, F and G.");
print ("    Number to be tested has ", length(Str(n)), " digits.");
print ("    Modulus has ", length(Str(s)), " digits.");
if ((n-1)%F != 0, \
    print ("WARNING:  Some factors of F are not factors of n-1.\n    Proceeding with a smaller value of F."));
if ((n+1)%G != 0, \
    print ("WARNING:  Some factors of G are not factors of n+1.\n    Proceeding with a smaller value of G."));
print ("Modulus is ", precision(100*log(s)/log(n), 1), "% of n.\n");

if (s^4 < n, \
    print ("F and G are not big enough.  Do more factorization.");quit(););

print ("NOTICE: This program assumes that n has passed");
print ("    a BLS PRP-test with n, F, and G as given.  If");
print ("    not, then any results will be invalid!\n");

rightendpoint = findrightendpoint (n, F, G);
leftendpoint = 1; \\findleftendpoint (n, F, G);

\\ It is possible that F and G are large enough that CHG is completely
\\ unnecessary.  We check for a few of these common cases.
if (rightendpoint == 0, \
    print ("CHG not required."); \
    print ("\nCongratulations! n is prime!"); \
    quit());
if (F^10 > n^3 || G^10 > n^3, \
    print ("\nWARNING:  This number qualifies for the KP test.  CHG may be unnecessary."));
if (s^3 > n, \
    print ("\nWARNING:  This number qualifies for Lenstra's algorithm.  CHG may be unnecessary."));

print ("\nSearch for factors congruent to 1.");
possiblefactor1 = CHGFactor(n, s, 1, leftendpoint, rightendpoint);
if (type(possiblefactor1) == "t_POL", print ("Factor found:  ", possiblefactor1);print ("n is composite.");quit());

\\ If G=1, the second CHG step is unnecessary and we are done.
\\   A certificate with no fourth element is saved.
\\    There is no corresponding speedup for F=1.
if (G == 1, print ("\nCongratulations! n is prime!"); certificateSave ([n, s, possiblefactor1]); quit());

print ("\nSearch for factors congruent to n.");
possiblefactor2 = CHGFactor(n, s, n%s, leftendpoint, rightendpoint);
if (type(possiblefactor2) == "t_POL", print ("Factor found:  ", possiblefactor2);print ("n is composite.");quit());

print ("\nCongratulations! n is prime!");
certificateSave ([n, s, possiblefactor1, possiblefactor2]);

quit();
